#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    :platform: Unix
    :synopsis: Processing context used in this module.

"""

from ESGConfigParser import SectionParser

from nctime.utils.collector import Collector
from nctime.utils.constants import *
from nctime.utils.context import BaseContext
from nctime.utils.custom_exceptions import NoConfigCardFound, NoRunCardFound
from nctime.utils.custom_print import *


class ProcessingContext(BaseContext):
    """
    Encapsulates the processing context/information for main process.

    :param ArgumentParser args: Parsed command-line arguments
    :returns: The processing context
    :rtype: *ProcessingContext*

    """

    def __init__(self, args):
        super(self.__class__, self).__init__(args)
        self.directory = args.directory
        self.resolve = args.resolve
        self.full_only = args.full_only
        self.overlaps = 0
        self.broken = 0
        self.card = args.card
        if args.card:
            self.card = list(yield_xml_from_card(args.card))
        self.nbnodes = 0
        self.nbdsets = 0

    def __enter__(self):
        # Init data collector
        self.sources = Collector(sources=self.directory)
        # Run __enter__() BaseContext
        super(self.__class__, self).__enter__()
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        msg = COLORS.OKBLUE('Number of node(s): {}\n'.format(self.nbnodes))
        msg += COLORS.HEADER('Number of dataset(s): {}\n'.format(self.nbdsets))
        m = 'Number of dataset(s) with overlap(s): {}\n'.format(self.overlaps)
        if self.overlaps:
            msg += COLORS.FAIL(m)
        else:
            msg += COLORS.SUCCESS(m)
        m = 'Number of dataset(s) with broken time series: {}'.format(self.broken)
        if self.broken:
            msg += COLORS.FAIL(m)
        else:
            msg += COLORS.SUCCESS(m)
        Print.summary(msg)
        # Run __exit__() BaseContext
        super(self.__class__, self).__exit__(exc_type, exc_val, traceback)


def yield_xml_from_card(card_path):
    """
    Yields XML path from run.card and config.card attributes.

    :param str card_path: Directory including run.card and config.card
    :returns: The XML paths to use
    :rtype: *iter*

    """
    # Check cards exist
    if RUN_CARD not in os.listdir(card_path):
        raise NoRunCardFound(card_path)
    else:
        run_card = os.path.join(card_path, RUN_CARD)
    if CONF_CARD not in os.listdir(card_path):
        raise NoConfigCardFound(card_path)
    else:
        conf_card = os.path.join(card_path, CONF_CARD)
    # Extract config info from config.card
    config = SectionParser('UserChoices')
    config.read(conf_card)
    xml_attrs = dict()
    xml_attrs['root'] = FILEDEF_ROOT
    xml_attrs['longname'] = config.get('longname').strip('"')
    xml_attrs['experimentname'] = config.get('experimentname').strip('"')
    if config.has_option('modelname'):
        xml_attrs['modelname'] = config.get('modelname').strip('"')
    else:
        xml_attrs['modelname'] = 'IPSL-CM6A-LR'
    xml_attrs['member'] = config.get('member').strip('"')
    # Extract first and last simulated years from run.card
    with open(run_card, 'r') as f:
        lines = f.read().split('\n')
    # Get run table without header
    lines = [line for line in lines if line.count('|') == 8][1:]
    year_start = int(lines[0].split()[3][:4])
    year_end = int(lines[-1].split()[5][:4])
    for year in range(year_start, year_end + 1):
        xml_attrs['year'] = str(year)
        yield FILEDEF_DIRECTORY_FORMAT.format(**xml_attrs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    :platform: Unix
    :synopsis: Class and methods used to parse command-line arguments.

"""

import os
import re
import sys
import textwrap
from argparse import HelpFormatter, ArgumentTypeError, Action, ArgumentParser
from gettext import gettext

from netcdftime import utime

from constants import TIME_UNITS, FREQ_INC, CALENDARS, TIME_UNITS_FORMAT
from custom_exceptions import InvalidUnits, InvalidFrequency, InvalidTable


class CustomParser(ArgumentParser):
    def error(self, message):
        """
        Overwrite the original method to change exist status.

        """
        self.print_usage(sys.stderr)
        self.exit(-1, gettext('%s: error: %s\n') % (self.prog, message))


class MultilineFormatter(HelpFormatter):
    """
    Custom formatter class for argument parser to use with the Python
    `argparse <https://docs.python.org/2/library/argparse.html>`_ module.

    """

    def __init__(self, prog):
        # Overload the HelpFormatter class.
        super(MultilineFormatter, self).__init__(prog, max_help_position=60, width=100)

    def _fill_text(self, text, width, indent):
        # Rewrites the _fill_text method to support multiline description.
        text = self._whitespace_matcher.sub(' ', text).strip()
        multiline_text = ''
        paragraphs = text.split('|n|n ')
        for paragraph in paragraphs:
            lines = paragraph.split('|n ')
            for line in lines:
                formatted_line = textwrap.fill(line, width,
                                               initial_indent=indent,
                                               subsequent_indent=indent) + '\n'
                multiline_text += formatted_line
            multiline_text += '\n'
        return multiline_text

    def _split_lines(self, text, width):
        # Rewrites the _split_lines method to support multiline helps.
        text = self._whitespace_matcher.sub(' ', text).strip()
        lines = text.split('|n ')
        multiline_text = []
        for line in lines:
            multiline_text.append(textwrap.fill(line, width))
        multiline_text[-1] += '\n'
        return multiline_text


class DirectoryChecker(Action):
    """
    Custom action class for argument parser to use with the Python
    `argparse <https://docs.python.org/2/library/argparse.html>`_ module.

    """

    def __call__(self, parser, namespace, values, option_string=None):
        if isinstance(values, list):
            checked_vals = [self.directory_checker(x) for x in values]
        else:
            checked_vals = self.directory_checker(values)
        setattr(namespace, self.dest, checked_vals)

    @staticmethod
    def directory_checker(path):
        """
        Checks if the supplied directory exists. The path is normalized without trailing slash.

        :param str path: The path list to check
        :returns: The same path list
        :rtype: *str*
        :raises Error: If one of the directory does not exist

        """
        path = os.path.abspath(os.path.normpath(path))
        if not os.path.isdir(path):
            msg = 'No such directory: {}'.format(path)
            raise ArgumentTypeError(msg)
        return path


class InputChecker(Action):
    """
    Checks if the supplied input exists.

    """

    def __call__(self, parser, namespace, values, option_string=None):
        checked_values = [self.input_checker(x) for x in values]
        setattr(namespace, self.dest, checked_values)

    @staticmethod
    def input_checker(path):
        path = os.path.abspath(os.path.normpath(path))
        if not os.path.exists(path):
            msg = 'No such input: {}'.format(path)
            raise ArgumentTypeError(msg)
        return path


class CodeChecker(Action):
    """
    Checks if the supplied input exists.

    """

    def __call__(self, parser, namespace, values, option_string=None):
        checked_values = self.code_checker(values)
        setattr(namespace, self.dest, checked_values)

    @staticmethod
    def code_checker(codes):
        allowed_codes = [str(x).rjust(3, '0') for x in range(0, 9)]
        if not isinstance(codes, str):
            msg = 'Type not supported: {} -- Should be a comma-separated string without spaces.'.format(codes)
            raise ArgumentTypeError(msg)
        codes = codes.split(',')
        for code in codes:
            if code not in allowed_codes:
                msg = 'Invalid code: {} -- Available codes are {}'.format(code, ', '.join(allowed_codes))
                raise ArgumentTypeError(msg)
        return codes


class TimestampChecker(Action):
    """
    Checks if the supplied timestamp is valid.

    """

    def __call__(self, parser, namespace, values, option_string=None):
        checked_values = self.timestamp_checker(values)
        setattr(namespace, self.dest, checked_values)

    @staticmethod
    def timestamp_checker(timestamp):
        if not timestamp.isdigit():
            msg = 'Bad timestamp format -- Digits only (e.g.: YYYYMMDDmmhhss).'
            raise ArgumentTypeError(msg)
        return timestamp


class CalendarChecker(Action):
    """
    Checks if the supplied calendar is valid.

    """

    def __call__(self, parser, namespace, values, option_string=None):
        checked_values = self.calendar_checker(values)
        setattr(namespace, self.dest, checked_values)

    @staticmethod
    def calendar_checker(calendar):
        if calendar not in CALENDARS:
            msg = 'Invalid calendar - Available calendars are {}'.format(', '.join(CALENDARS))
            raise ArgumentTypeError(msg)
        return calendar


class TimeUnitsChecker(Action):
    """
    Checks if the supplied time units has valid format.

    """

    def __call__(self, parser, namespace, values, option_string=None):
        checked_values = self.units_checker(values)
        setattr(namespace, self.dest, checked_values)

    @staticmethod
    def units_checker(units):
        try:
            u = utime(units)
            return u.unit_string
        except ValueError, TypeError:
            msg = 'Invalid time units format - Available format is "{}"'.format(TIME_UNITS_FORMAT)
            raise ArgumentTypeError(msg)



def regex_validator(string):
    """
    Validates a Python regular expression syntax.

    :param str string: The string to check
    :returns: The Python regex
    :rtype: *re.compile*
    :raises Error: If invalid regular expression

    """
    try:
        return re.compile(string)
    except re.error:
        msg = 'Bad regex syntax: {}'.format(string)
        raise ArgumentTypeError(msg)


def positive_only(value):
    """
    Validates a positive number.

    :param str value: The value submitted
    :return:
    """
    value = int(value)
    if value < 0:
        msg = 'Invalid number. Should be a positive integer.'
        raise ArgumentTypeError(msg)
    return value


def processes_validator(value):
    """
    Validates the max processes number.

    :param str value: The max processes number submitted
    :return:
    """
    pnum = int(value)
    if pnum < 1 and pnum != -1:
        msg = 'Invalid processes number. Should be a positive integer or "-1".'
        raise ArgumentTypeError(msg)
    if pnum == -1:
        # Max processes = None corresponds to cpu.count() in Pool creation
        return None
    else:
        return pnum


def inc_converter(string):
    """
    Checks the increment value syntax.

    :param str string: The string to check
    :returns: The key/value tuple
    :rtype: *list*
    :raises Error: If invalid pair syntax

    """
    pattern = re.compile(r'([^=]+):([^=]+)=([^=]+)(?:,|$)')
    if not pattern.search(string):
        msg = 'Bad argument syntax: {}'.format(string)
        raise ArgumentTypeError(msg)
    else:
        table, frequency, v = pattern.search(string).groups()
        inc, units = v[:-1], v[-1]
        if not inc.isdigit():
            msg = 'Bad argument syntax -- only digits allowed: {}'.format(inc)
            raise ArgumentTypeError(msg)
        tables = set(zip(*FREQ_INC.keys())[0])
        tables.add('all')
        if table not in tables:
            raise InvalidTable(table)
        frequencies = set(zip(*FREQ_INC.keys())[1])
        frequencies.add('all')
        if frequency not in frequencies:
            raise InvalidFrequency(frequency)
        if units not in TIME_UNITS.keys():
            raise InvalidUnits(units)
        return table, frequency, inc, TIME_UNITS[units]


def table_inc_converter(pair):
    """
    Checks the key value syntax.

    :param str pair: The key/value pair to check
    :returns: The key/value pair
    :rtype: *list*
    :raises Error: If invalid pair syntax

    """
    pattern = re.compile(r'([^=]+)=([^=]+)(?:,|$)')
    if not pattern.search(pair):
        msg = 'Bad argument syntax: {}'.format(pair)
        raise ArgumentTypeError(msg)
    else:
        table, v = pattern.search(pair).groups()
        inc, units = v[:-1], v[-1]
        if not inc.isdigit():
            msg = 'Bad argument syntax -- only digits allowed: {}'.format(inc)
            raise ArgumentTypeError(msg)
        if table not in TABLE_INC.keys():
            raise InvalidTable(table)
        if units not in TIME_UNITS.keys():
            raise InvalidUnits(units)
        return table, inc, TIME_UNITS[units]



#!/usr/bin/env python
"""
   :platform: Unix
   :synopsis: File handler for time axis.

"""

from copy import deepcopy as copy

import nco
import numpy as np
from fuzzywuzzy import fuzz, process

from constants import *
from custom_exceptions import *
from nctime.utils.constants import CLIM_SUFFIX, AVERAGE_CORRECTION_FREQ
from nctime.utils.custom_exceptions import *
from nctime.utils.custom_print import *
from nctime.utils.misc import ncopen
from nctime.utils.time import truncated_timestamp, get_start_end_dates_from_filename, dates2str, num2date, date2num, \
    control_time_units, trunc, time_inc, convert_time_units


class File(object):
    """
    Handler providing methods to deal with file processing.

    :returns: The file handler
    :rtype: *File*

    """

    def __init__(self, ffp, pattern, ref_units, ref_calendar, input_start_timestamp=None, input_end_timestamp=None):
        # Retrieve the file full path
        self.ffp = ffp
        # Retrieve the reference time units to use
        self.ref_units = ref_units
        # Retrieve the reference calendar to use
        self.ref_calendar = ref_calendar
        # Retrieve the file size
        self.size = os.stat(self.ffp).st_size
        # Retrieve directory and filename full path
        self.directory, self.filename = os.path.split(ffp)
        # Remove "-clim.nc" suffix from filename if exists
        self.name = self.filename.replace(CLIM_SUFFIX, '.nc') if self.filename.endswith(CLIM_SUFFIX) else self.filename
        # Set variables for time axis diagnostic
        self.time_axis_rebuilt = None
        self.date_axis_rebuilt = None
        self.time_bounds_rebuilt = None
        self.date_bounds_rebuilt = None
        self.status = list()
        # Get table from file
        try:
            self.table = self.nc_att_get('table_id')
            # Extract MIP table from string if needed
            self.table = self.table.split(" ")[1]
        except IndexError:
            self.table = self.nc_att_get('table_id')
        except NoNetCDFAttribute:
            self.table = 'None'
        # Rollback to None if unknown table
        if self.table not in set(zip(*FREQ_INC.keys())[0]):
            msg= 'Unknown MIP table "{}" -- Consider default increment for the given frequency.'.format(self.table)
            Print.warning(msg)
            self.table = 'None'
        # Get frequency from file
        self.frequency = self.nc_att_get('frequency')
        # Get netCDF time properties
        with ncopen(self.ffp) as nc:
            # Get time length and vector
            if 'time' not in nc.variables.keys():
                raise NoNetCDFVariable('time', self.ffp)
            self.length = nc.variables['time'].shape[0]
            if self.length == 0:
                raise EmptyTimeAxis(self.ffp)
            t = nc.variables['time'][:]
            self.time_axis = trunc(t, NDECIMALS)
            self.date_axis = dates2str(num2date(t, units=self.ref_units, calendar=self.ref_calendar))
            # Get time boundaries
            self.has_bounds = False
            self.time_bounds = None
            self.date_bounds = None
            self.tbnds = None
            if 'bounds' in nc.variables['time'].ncattrs():
                self.has_bounds = True
                self.tbnds = nc.variables['time'].bounds
            if 'climatology' in nc.variables['time'].ncattrs():
                self.has_bounds = True
                self.tbnds = nc.variables['time'].climatology
            if self.tbnds:
                bnds = nc.variables[self.tbnds][:, :]
                self.time_bounds = trunc(bnds, NDECIMALS)
                self.date_bounds = np.column_stack((
                    dates2str(num2date(bnds[:, 0], units=self.ref_units, calendar=self.ref_calendar)),
                    dates2str(num2date(bnds[:, 1], units=self.ref_units, calendar=self.ref_calendar))
                ))
                del bnds
            # Get time units from file
            if 'units' not in nc.variables['time'].ncattrs():
                raise NoNetCDFAttribute('units', self.ffp, 'time')
            self.tunits = control_time_units(nc.variables['time'].units)
            # Get calendar from file
            if 'calendar' not in nc.variables['time'].ncattrs():
                raise NoNetCDFAttribute('calendar', self.ffp, 'time')
            self.calendar = nc.variables['time'].calendar
            # Get boolean on instantaneous time axis
            variable = unicode(self.filename.split('_')[0])
            if 'cell_methods' not in nc.variables[variable].ncattrs():
                raise NoNetCDFAttribute('cell_methods', self.ffp, variable)
            self.is_instant = False
            if 'time: point' in nc.variables[variable].cell_methods.lower():
                self.is_instant = True
            # Get boolean on climatology time axis
            self.is_climatology = False
            if 'climatology' in nc.variables['time'].ncattrs():
                self.is_climatology = True
        # Get time step increment from frequency and table
        self.step, self.step_units = time_inc(self.table, self.frequency)
        # Convert reference time units into frequency units depending on the file (i.e., months/year/hours since ...)
        self.funits = convert_time_units(self.ref_units, self.table, self.frequency)
        # Get timestamps length from filename
        self.timestamp_length = len(re.match(pattern, self.name).groupdict()['period_end'])
        # Overwrite filename timestamp if submitted
        # Extract start and end dates from filename
        dates = get_start_end_dates_from_filename(filename=self.name,
                                                  pattern=pattern,
                                                  table=self.table,
                                                  frequency=self.frequency,
                                                  calendar=self.calendar,
                                                  start=input_start_timestamp,
                                                  end=input_end_timestamp)
        dates_num = trunc(date2num(dates, units=self.funits, calendar=self.calendar), NDECIMALS)
        if self.is_climatology:
            # Get climatology offset to start in the middle of the interval
            year_diff = dates[1].year - dates[0].year
            start_clim_date = dates[0].replace(year=dates[0].year + year_diff / 2)
            end_clim_date = dates[1].replace(year=dates[0].year + year_diff / 2)
            start_clim_num = trunc(date2num(start_clim_date, units=self.funits, calendar=self.calendar), NDECIMALS)
            end_clim_num = trunc(date2num(end_clim_date, units=self.funits, calendar=self.calendar), NDECIMALS)
            # Apply time offset corresponding to the climatology:
            self.clim_diff = [start_clim_num - dates_num[0], dates_num[1] - end_clim_num]
            if self.frequency in ['monC', 'monClim']:
                self.clim_diff.append(dates_num[1] - 10 - start_clim_num)
            elif self.frequency == '1hrCM':
                self.clim_diff.append(dates_num[1] - 22.5 - start_clim_num)
            else:
                raise InvalidClimatologyFrequency(self.frequency)
            dates_num[0] += self.clim_diff[0] + 0.5
            dates_num[1] -= self.clim_diff[1] - 0.5
        elif not self.is_instant and self.frequency in AVERAGE_CORRECTION_FREQ:
            # Apply time offset for non-instant time axis:
            dates_num += 0.5
        self.start_axis = dates_num[0]
        dates = num2date(dates_num, units=self.funits, calendar=self.calendar)
        self.start_num, self.end_num, _ = date2num(dates, units=self.tunits, calendar=self.calendar)
        self.start_date, self.end_date, _ = dates2str(list(dates))
        # Convert dates into timestamps
        self.start_timestamp, self.end_timestamp, _ = [
            truncated_timestamp(date, self.timestamp_length) for date in dates]
        # Declare last time attribute
        self.last_date = None
        self.last_timestamp = None
        self.last_num = None

    def build_time_axis(self):
        """
        Rebuilds time axis from date axis, depending on MIP frequency, calendar and instant status.

        :returns: The corresponding theoretical time axis
        :rtype: *numpy.array*

        """
        num_axis = np.arange(start=self.start_axis,
                             stop=self.start_axis + self.length * self.step,
                             step=self.step)
        num_axis = self.check_axis_length(num_axis)
        date_axis = num2date(num_axis, units=self.funits, calendar=self.ref_calendar)
        del num_axis
        axis_rebuilt = date2num(date_axis, units=self.ref_units, calendar=self.ref_calendar)
        self.date_axis_rebuilt = dates2str(num2date(axis_rebuilt, units=self.ref_units, calendar=self.ref_calendar))
        return axis_rebuilt

    def build_time_bounds(self):
        """
        Rebuilds time boundaries from the start date, depending on MIP frequency, calendar and
        instant status.

        :returns: The corresponding theoretical time boundaries as a [n, 2] array
        :rtype: *numpy.array*

        """
        num_axis = np.arange(start=self.start_axis,
                             stop=self.start_axis + self.length * self.step,
                             step=self.step)
        num_axis = self.check_axis_length(num_axis)
        num_axis_bnds_inf, num_axis_bnds_sup = num_axis, copy(num_axis)
        if self.is_climatology:
            num_axis_bnds_inf -= self.clim_diff[0] + 0.5
            num_axis_bnds_sup += self.clim_diff[2] - 0.5
        elif not self.is_instant:
            num_axis_bnds_inf -= 0.5 * self.step
            num_axis_bnds_sup += 0.5 * self.step
        num_axis_bnds = np.column_stack((num_axis_bnds_inf, num_axis_bnds_sup))
        del num_axis, num_axis_bnds_inf, num_axis_bnds_sup
        date_axis_bnds = num2date(num_axis_bnds, units=self.funits, calendar=self.ref_calendar)
        del num_axis_bnds
        axis_bnds_rebuilt = date2num(date_axis_bnds, units=self.ref_units, calendar=self.ref_calendar)
        self.date_bounds_rebuilt = np.column_stack((
            dates2str(num2date(axis_bnds_rebuilt[:, 0], units=self.ref_units, calendar=self.ref_calendar)),
            dates2str(num2date(axis_bnds_rebuilt[:, 1], units=self.ref_units, calendar=self.ref_calendar))
        ))
        return axis_bnds_rebuilt

    def check_axis_length(self, axis):
        """
        numpy.arange could suddenly add last endpoint to the array in the case of high number of steps
        Due to rounding float issue and pre-calculated length in memory.
        As a workaround, always check length and remove last point if length are different.

        :param *numpy.array* axis: The axis to check
        :returns: The cut axis
        :rtype: *numpy.array*

        """
        if len(axis) == self.length + 1:
            axis = axis[:-1]
        assert len(axis) == self.length
        return axis

    def nc_var_delete(self, variable):
        """
        Delete a NetCDF variable using NCO operators.
        A unique filename is generated to avoid multiprocessing errors.
        To overwrite the input file, the source file is dump using the ``cat`` Shell command-line
        to avoid Python memory limit.

        :param str variable: The variable to delete
        :raises Error: If the deletion failed

        """
        try:
            nc = nco.Nco()
            nc.ncks(input=self.ffp,
                    options=['-O', '-x', '-v {}'.format(variable)])
        except:
            raise NetCDFVariableRemoveFail(variable, self.ffp)

    def nc_att_delete(self, variable, attribute):
        """
        Delete a NetCDF dimension attribute using NCO operators.
        A unique filename is generated to avoid multiprocessing errors.
        To overwrite the input file, the source file is dump using the ``cat`` Shell command-line
        to avoid Python memory limit.

        :param str attribute: The attribute to delete
        :param str variable: The variable that has the attribute
        :raises Error: If the deletion failed

        """
        try:
            nc = nco.Nco()
            nc.ncatted(input=self.ffp,
                       options=['-O', '-a {},{},d,,'.format(attribute, variable)])
        except:
            raise NetCDFAttributeRemoveFail(attribute, self.ffp, variable)

    def nc_var_overwrite(self, variable, data):
        """
        Rewrite variable to NetCDF file without copy.

        :param str variable: The variable to replace
        :param float array data: The data array to overwrite

        """
        with ncopen(self.ffp, 'r+') as nc:
            nc.variables[variable][:] = data

    def nc_att_overwrite(self, attribute, data, variable=None):
        """
        Rewrite attribute to NetCDF file without copy.

        :param str attribute: The attribute to replace
        :param str data: The string to add to overwrite
        :param str variable: The variable that has the attribute, default is global attributes

        """
        with ncopen(self.ffp, 'r+') as nc:
            if variable:
                if variable not in nc.variables.keys():
                    raise NoNetCDFVariable(variable, nc.path)
                nc.variables[variable].setncattr(attribute, data)
            else:
                nc.setncattr(attribute, data)

    def nc_att_get(self, attribute, variable=None):
        """
        Get attribute from NetCDF file. Default is to find into global attributes.
        If attribute key is not found, get the closest key name instead.


        :param str attribute: The attribute key to get
        :param str variable: The variable from which to find the attribute. Global is None.
        :return: The attribute value
        :rtype: *str*

        """
        with ncopen(self.ffp) as nc:
            if variable:
                attrs = nc.variables[variable].__dict__
            else:
                attrs = nc.__dict__
            if attribute in attrs.keys():
                return attrs[attribute]
            else:
                key, score = process.extractOne(attribute, attrs, scorer=fuzz.partial_ratio)
                if score >= 80:
                    Print.warning('Consider "{}" attribute instead of "frequency"'.format(key))
                    return attrs(key)
                else:
                    raise NoNetCDFAttribute(attribute, self.ffp)

    def nc_file_rename(self, new_filename):
        """
        Rename a NetCDF file.

        :param str new_filename: The new filename to apply

        """
        ffp = os.path.join(os.path.dirname(self.ffp), new_filename)
        if os.path.exists(ffp):
            raise RenamingNetCDFFailed(self.ffp, ffp, exists=True)
        try:
            os.rename(self.ffp, ffp)
        except OSError:
            raise RenamingNetCDFFailed(self.ffp, ffp)
        self.ffp = ffp
        self.filename = new_filename



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
   :platform: Unix
   :synopsis: Useful functions to use with this package.

"""

from fuzzywuzzy import fuzz, process
from netCDF4 import Dataset

from custom_exceptions import *
from custom_print import *


class ncopen(object):
    """
    Properly opens a netCDF file

    :param str path: The netCDF file full path
    :returns: The netCDF dataset object
    :rtype: *netCDF4.Dataset*

    """

    def __init__(self, path, mode='r'):
        self.path = path
        self.mode = mode
        self.nc = None

    def __enter__(self):
        try:
            self.nc = Dataset(self.path, self.mode)
        except IOError:
            raise InvalidNetCDFFile(self.path)
        return self.nc

    def __exit__(self, *exc):
        self.nc.close()


def match(pattern, string, inclusive=True):
    """
    Validates a string against a regular expression.
    Only match at the beginning of the string.
    Default is to match inclusive regex.

    :param str pattern: The regular expression to match
    :param str string: The string to test
    :param boolean inclusive: False if negative matching (i.e., exclude the regex)
    :returns: True if it matches
    :rtype: *boolean*

    """
    # Assert inclusive and exclusive flag are mutually exclusive
    if inclusive:
        return True if re.search(pattern, string) else False
    else:
        return True if not re.search(pattern, string) else False


def get_project(ffp):
    """
    Get project identifier from netCDF file.

    :param str ffp: The file full path
    :returns: The lower-case project id
    :rtype: *str*

    """
    project = None
    with ncopen(ffp) as nc:
        if 'mip_era' in nc.ncattrs():
            project = nc.getncattr('mip_era')
        elif 'project' in nc.ncattrs():
            project = nc.getncattr('project')
        else:
            key, score = process.extractOne('project', nc.ncattrs(), scorer=fuzz.partial_ratio)
            if score >= 80:
                project = nc.getncattr(key)
                Print.warning('Consider "{}" attribute instead of "project"'.format(key))
            else:
                raise NoNetCDFAttribute('project', ffp)
    return project.lower()


class ProcessContext(object):
    """
    Encapsulates the processing context/information for child process.

    :param dict args: Dictionary of argument to pass to child process
    :returns: The processing context
    :rtype: *ProcessContext*

    """

    def __init__(self, args):
        assert isinstance(args, dict)
        for key, value in args.items():
            setattr(self, key, value)



#!/usr/bin/env python
"""
   :platform: Unix
   :synopsis: File handler for time axis.

"""

from copy import deepcopy as copy

from fuzzywuzzy import fuzz, process
from networkx import DiGraph

from nctime.utils.constants import CLIM_SUFFIX
from nctime.utils.custom_exceptions import *
from nctime.utils.custom_print import *
from nctime.utils.misc import ncopen
from nctime.utils.time import get_start_end_dates_from_filename, dates2int


class Filename(object):
    """
    Handler providing methods to deal with file processing.

    """

    def __init__(self, ffp):
        self.ffp = ffp
        # Retrieve filename
        self.filename = os.path.basename(ffp)
        # Remove "-clim.nc" suffix from filename if exists
        self.name = self.filename.replace(CLIM_SUFFIX, '.nc') if self.filename.endswith(CLIM_SUFFIX) else self.filename
        # Build id as the filename without the dates and extension
        if self.filename.endswith(CLIM_SUFFIX):
            self.id = '_'.join(self.filename.split('_')[:-1]) + '-clim'
        else:
            self.id = '_'.join(self.filename.split('_')[:-1])
        # Get first and last time steps
        # Start/end period dates from filename + next expected date
        self.start_date = None
        self.end_date = None
        self.next_date = None

    def get_start_end_dates(self, pattern, calendar):
        """
        Wraps and records :func:`get_start_end_dates_from_filename` results.

        :param re Object pattern: The filename pattern as a regex (from `re library \
        <https://docs.python.org/2/library/re.html>`_).
        :param str calendar: The NetCDF calendar attribute
        :returns: Start and end dates as number of days since the referenced date
        :rtype: *float*

        """
        # Get table from file
        try:
            table = self.nc_att_get('table_id')
            # Extract MIP table from string if needed
            table = table.split(" ")[1]
        except IndexError:
            table = self.nc_att_get('table_id')
        except NoNetCDFAttribute:
            table = 'None'
        # Rollback to None if unknown table
        if table not in set(zip(*FREQ_INC.keys())[0]):
            msg = 'Unknown MIP table "{}" -- Consider default increment for the given frequency.'.format(table)
            Print.warning(msg)
            table = 'None'
        # Get frequency from file
        frequency = self.nc_att_get('frequency')
        dates = get_start_end_dates_from_filename(filename=self.name,
                                                  pattern=pattern,
                                                  table=table,
                                                  frequency=frequency,
                                                  calendar=calendar)
        self.start_date, self.end_date, self.next_date = dates2int(dates)

    def nc_att_get(self, attribute, variable=None):
        """
        Get attribute from NetCDF file. Default is to find into global attributes.
        If attribute key is not found, get the closest key name instead.


        :param str attribute: The attribute key to get
        :param str variable: The variable from which to find the attribute. Global is None.
        :return: The attribute value
        :rtype: *str*

        """
        with ncopen(self.ffp) as nc:
            if variable:
                attrs = nc.variables[variable].__dict__
            else:
                attrs = nc.__dict__
            if attribute in attrs.keys():
                return attrs[attribute]
            else:
                key, score = process.extractOne(attribute, attrs, scorer=fuzz.partial_ratio)
                if score >= 80:
                    Print.warning('Consider "{}" attribute instead of "frequency"'.format(key))
                    return attrs(key)
                else:
                    raise NoNetCDFAttribute(attribute, self.ffp)


class Graph(object):
    """
    Handler providing methods to deal with Directed graph.

    """

    def __init__(self):
        pass

    def has_graph(self, i):
        return True if hasattr(self, i) else False

    def set_graph(self, i, g=DiGraph()):
        setattr(self, i, copy(g))

    def get_graph(self, i):
        return copy(getattr(self, i))

    def add_node(self, i, filename, start, end, next, path):
        g = getattr(self, i)
        g.add_node(filename,
                   start=start,
                   end=end,
                   next=next,
                   path=path)
        return g.node[filename]

    def add_edge(self, i, node_src, node_dst):
        g = getattr(self, i)
        g.add_edge(node_src, node_dst)

    def __call__(self, *args, **kwargs):
        return [id for id in self.__dict__]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    :platform: Unix
    :synopsis: Processing context used in this module.

"""

from ESGConfigParser import SectionParser

from nctime.utils.collector import Collector
from nctime.utils.constants import *
from nctime.utils.context import BaseContext
from nctime.utils.custom_exceptions import NoRunCardFound
from nctime.utils.custom_print import *


class ProcessingContext(BaseContext):
    """
    Encapsulates the processing context/information for main process.

    :param ArgumentParser args: Parsed command-line arguments
    :returns: The processing context
    :rtype: *ProcessingContext*

    """

    def __init__(self, args):
        super(self.__class__, self).__init__(args)
        self.input = args.input
        self.write = args.write
        self.force = args.force
        self.debug = args.debug
        self.ref_start = args.start
        self.ref_end = args.end
        self.on_fly = args.on_fly
        if args.card:
            self.on_fly = True if not is_simulation_completed(args.card) else False
        self.limit = args.limit
        self.ignore_codes = args.ignore_errors
        self.status = []

    def __enter__(self):
        # Print warning message if on-fly mode
        if self.on_fly:
            Print.warning('"on-fly" mode activated -- Incomplete time axis expected')
        # Init data collector
        self.sources = Collector(sources=self.input, spinner=False)
        # Run __enter__() BaseContext
        super(self.__class__, self).__enter__()
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        m = 'Number of file(s) with error(s): {}'.format(self.nberrors)
        if self.nberrors:
            msg = COLORS.FAIL(m)
        else:
            msg = COLORS.SUCCESS(m)
        Print.summary(msg)
        # Run __exit__() BaseContext
        super(self.__class__, self).__exit__(exc_type, exc_val, traceback)


def is_simulation_completed(card_path):
    """
    Returns True if the simulation is completed.

    :param str card_path: Directory including run.card
    :returns: True if the simulation is completed
    :rtype: *boolean*

    """
    # Check cards exist
    if RUN_CARD not in os.listdir(card_path):
        raise NoRunCardFound(card_path)
    else:
        run_card = os.path.join(card_path, RUN_CARD)
    # Extract info from cards
    config = SectionParser('Configuration')
    config.read(run_card)
    return config.get('periodstate').strip('"') == 'Completed'
